"""Lossless compression experiment using enb, the Electronic NoteBook library.
See https://github.com/miguelinux314/experiment-notebook for additional documentation.

This script installs the plugins needed to perform compression experiments (some of them need to be gathered by the user due to licensing restrictions),
performs the compression experimnets with the selected plugins on the dataset obtained from the ../RAW folder, and performs the compression analysis.

Results obtained can be found in analysis/ScalarNumericAnalyzer-6columns-histogram-groupby__dataset_name__task_label.csv for all each dataset results
and in analysis/ScalarNumericAnalyzer-6columns-histogram-groupby__task_label for all corpus average results
"""
__author__ = "Òscar Maireles-González and Miguel Hernández-Cabronero"
__since__ = "2022/07/19"

import os
import sys
import subprocess
import enb
from enb.config import options
from enb import aanalysis, icompression

print("The following packages need to be installed (e.g., with apt) in order to be able to install all needed plugins:"+"\n"+"\n"+\
    "cmake"+"\n"+"libboost-dev"+"\n"+"libboost-thread-dev"+"\n"+"libboost-program-options-dev"+"\n"+"libboost-system-dev"+"\n")

plugin_list=["jpeg", "fpack", "kakadu", "lcnl", "hevc", "speck", "zip", "hdf5", "fpc", "spdp", \
            "lz4", "ndzip", "fapec", "fpzip", "zfp", "zstd"]
for plugin in plugin_list:
    enb.plugins.install(plugin, target_dir=f"plugins/plugin_{plugin}")

import plugins.plugin_jpeg.jpeg_codecs
import plugins.plugin_fpack.fpack_codec
import plugins.plugin_fapec.fapec_codec
import plugins.plugin_kakadu.kakadu_codec
import plugins.plugin_lcnl.lcnl_codecs
import plugins.plugin_hevc.hevc_codec
import plugins.plugin_zip.zip_codecs
import plugins.plugin_fpzip.fpzip_codec
import plugins.plugin_zfp.zfp_codec
import plugins.plugin_zstd.zstd_codec
import plugins.plugin_hdf5.hdf5_codecs
import plugins.plugin_fpc.fpc_codec
import plugins.plugin_spdp.spdp_codec
import plugins.plugin_lz4.lz4_codec
import plugins.plugin_ndzip.ndzip_codec
import plugins.plugin_speck.speck_codec


# All plugins with codecs installed under plugins/ are automatically imported

class LosslessExperiment(enb.icompression.LosslessCompressionExperiment):
    """Custom data columns can be defined in this class.
    If no columns are defined, enb.icompression.LosslessCompressionExperiment can
    be used directly when instantiating the experiment.
    """
    def column_dataset_name(self, index, row):
        path, task = self.index_to_path_task(index)
        return os.path.basename(os.path.abspath(os.path.dirname(path)))

if __name__ == '__main__':

    # Temporary compressed and reconstructed files are saved under options.base_tmp_dir.
    # By default, a memory disk is used if possible. This can be set to a physical partition if
    # temporary files are expected to be larger than the memory (swap may be used).
    #options.base_tmp_dir = "tmp"

    # By default, Experiments and other ATable subclasses read input samples from options.base_dataset_dir,
    # which defaults to ./data.

    if os.path.exists('./data') and len(os.listdir('./data')) != 0:
        pass
    elif os.path.exists('./data') == False and os.path.exists('../RAW') == True and len(os.listdir('../RAW')) != 0:
        os.mkdir('./data')
        os.symlink(os.path.abspath('../RAW/INT/u16be'), os.path.abspath('./data/INT'))
        os.symlink(os.path.abspath('../RAW/JKT/u16be'), os.path.abspath('./data/JKT'))
        os.symlink(os.path.abspath('../RAW/LCO/u16be'), os.path.abspath('./data/LCO'))
        os.symlink(os.path.abspath('../RAW/TJO/u16be'), os.path.abspath('./data/TJO'))
        os.symlink(os.path.abspath('../RAW/WHT/u16be'), os.path.abspath('./data/WHT'))
    else:
        print('Dataset not found! Do you want to download astronomical dataset (1.5 GB)?')
        YesNo= str(input("Y/n: "))
        if YesNo == "Y":
            os.system('wget https://gici.uab.cat/datasets/AstronomicalData/astronomical_data_raw.zip -P ../')
            os.system('unzip ../astronomical_data_raw.zip -d ../')
            os.system('rm ../astronomical_data_raw.zip ')
            print('Dataset saved to ../RAW/')
            os.mkdir('./data')
            os.symlink(os.path.abspath('../RAW/INT/u16be'), os.path.abspath('./data/INT'))
            os.symlink(os.path.abspath('../RAW/JKT/u16be'), os.path.abspath('./data/JKT'))
            os.symlink(os.path.abspath('../RAW/LCO/u16be'), os.path.abspath('./data/LCO'))
            os.symlink(os.path.abspath('../RAW/TJO/u16be'), os.path.abspath('./data/TJO'))
            os.symlink(os.path.abspath('../RAW/WHT/u16be'), os.path.abspath('./data/WHT'))
        else:
            sys.exit("ERROR: data not found."+"\n"+"Astronomcial data can be downloaded from https://gici.uab.cat/GiciWebPage/datasets.php#astronomical")

    options.base_dataset_dir = './data'
    # Compression experiments use Codecs as ExperimentTask subclasses. A list of codec
    # instances (each with their desired parameters) needs to be created.
    # You can install new codecs (run `enb plugin list` to see a list of all available plugins)
    # into ./plugins and instantiate them at will.
    codecs = []
    if os.path.isfile('./plugins/plugin_kakadu/kdu_compress') == True:
        codecs.append(plugins.plugin_kakadu.kakadu_codec.Kakadu2DHaar(spatial_dwt_levels=9))
        codecs.append(plugins.plugin_kakadu.kakadu_codec.KakaduMCT(spatial_dwt_levels=9))
    else:
        print("Kakadu codec can not be used, binaries not found!"+"\n"+\
        "You will need to gather your own binaries (due to licensing restrictions) for the following plugins: Kakadu"+"\n"+\
        "Kakadu binaries can be gathered from https://kakadusoftware.com"+"\n"+\
        "Kakadu binaries 'kdu_compress' and 'kdu_expand' need to be placed in plugins/plugin_kakadu"+"\n")

    if os.path.isfile('./plugins/plugin_fapec/fapec') == True:
        codecs.append(plugins.plugin_fapec.fapec_codec.FAPEC_LP(linear_prediction_order = 3, chunk_size_str="8M", adaptiveness_block_length=512))
        codecs.append(plugins.plugin_fapec.fapec_codec.FAPEC_DWT(chunk_size_str="8M", adaptiveness_block_length=512))
        codecs.append(plugins.plugin_fapec.fapec_codec.FAPEC_NP(chunk_size_str="8M", adaptiveness_block_length=512))

    else:
        print("FAPEC codec can not be used, binaries not found!"+"\n"+\
        "You will need to gather your own binaries (due to licensing restrictions) for the following plugins: FAPEC"+"\n"+\
        "FAPEC binaries can be gathered from https://www.dapcom.es/get-fapec/"+"\n"+\
        "FAPEC binaries 'fapec' and 'unfapec' need to be placed in plugins/plugin_fapec"+"\n")

    if os.path.isfile('./plugins/plugin_lcnl/lcnl_encoder') == True:
        codecs.append(plugins.plugin_lcnl.lcnl_codecs.CCSDS_LCNL(entropy_coder_type=2, local_sum_type=0,prediction_mode=1))
    else:
        print("lcln codec can not be used, binaries not found!"+"\n"+\
        "You will need to gather your own binaries (due to licensing restrictions) for the following plugins: lcnl"+"\n"+\
        "lcnl binaries can be gathered from  https://logiciels.cnes.fr/en/license/173/728"+"\n"+\
        "lcnl binaries 'lcnl_encoder', 'lcnl_decoder' and 'lcnl_header_tool' need to be placed in plugins/plugin_kakadu"+"\n")

    codecs.append(plugins.plugin_jpeg.jpeg_codecs.JPEG_LS(max_error=0))
    codecs.append(plugins.plugin_fpack.fpack_codec.FPACK_RICE())
    codecs.append(plugins.plugin_fpack.fpack_codec.FPACK_HCOMPRESS())
    codecs.append(plugins.plugin_zstd.zstd_codec.Zstandard(compression_level=22))
    codecs.append(plugins.plugin_zstd.zstd_codec.Zstandard_train(compression_level=22))
    codecs.append(plugins.plugin_zip.zip_codecs.LZMA(compression_level=9))
    codecs.append(plugins.plugin_hevc.hevc_codec.HEVC_lossless(chroma_format="400"))
    codecs.append(plugins.plugin_zip.zip_codecs.BZIP2(compression_level=9))

    #codecs.append(plugins.plugin_hdf5.hdf5_codecs.SZIP(compression_level=9))
    #codecs.append(plugins.plugin_spdp.spdp_codec.Spdp(compression_level=9))
    #codecs.append(plugins.plugin_hdf5.hdf5_codecs.LZF(compression_level=9))
    #codecs.append(plugins.plugin_lz4.lz4_codec.Lz4(compression_level=9))
    #codecs.append(plugins.plugin_ndzip.ndzip_codec.Ndzip())#float
    #codecs.append(plugins.plugin_fpzip.plugin_fpzip.Fpzip())#float
    #codecs.append(plugins.plugin_zfp.plugin_zfp.Zfp())#float

    # Create experiment
    # All plugins installed in ./plugins can be instantiated automatically with the following snippet.

    if not codecs:
        enb.logger.error("No codec has been defined, nor installed under ./plugins. "
                         "Try `enb plugin list codec` for a list of "
                         "available plugins. You can change 'codec' for any other search string.")
        sys.exit(1)

    # Create experiment.
    exp = LosslessExperiment(codecs=codecs)



    # Generate pandas dataframe with results. Persistence is automatically added
    df = exp.get_df()

    # Selection of potentially relevant columns. You can use any of the columns available in
    # the experiment class (including any column you defined, e.g., in the LosslessExperiment class above).
    scalar_columns = ["compression_ratio_dr", "bpppc", "compression_time_seconds", "decompression_time_seconds",
                      "compression_efficiency_1byte_entropy", "compression_efficiency_2byte_entropy"]
    column_pairs = [("bpppc", "compression_time_seconds"),
                    ("compression_time_seconds", "compression_efficiency_1byte_entropy"),
                    ("compression_time_seconds", "compression_efficiency_2byte_entropy")]


    # Scalar column analysis for each dataset

    scalar_analyzer = enb.aanalysis.ScalarNumericAnalyzer(
        # A scalar analysis summary is stored here
        csv_support_path=os.path.join(options.analysis_dir, "lossless_compression_analysis_scalar_dataset.csv"))
    scalar_analyzer.get_df(
        # Mandatory params
        full_df=df,  # the dataframe produced by exp
        target_columns=scalar_columns,  # the list of ATable column names
        # Provide plotting hints based on the defined columns
        column_to_properties=exp.joined_column_to_properties,
        # Optional params
        #group_by="task_label" and "file_path",  # one can group by any column name
        group_by=["dataset_name", "task_label"],  # one can group by any column name
        selected_render_modes={"histogram"},
    )

    # Scalar column analysis for all corpus

    scalar_analyzer = enb.aanalysis.ScalarNumericAnalyzer(
        # A scalar analysis summary is stored here
        csv_support_path=os.path.join(options.analysis_dir, "lossless_compression_analysis_scalar_all.csv"))
    scalar_analyzer.get_df(
        # Mandatory params
        full_df=df,  # the dataframe produced by exp
        target_columns=scalar_columns,  # the list of ATable column names
        # Provide plotting hints based on the defined columns
        column_to_properties=exp.joined_column_to_properties,
        # Optional params
        #group_by="task_label" and "file_path",  # one can group by any column name
        group_by=["task_label"],  # one can group by any column name
        selected_render_modes={"histogram"},
    )

    # Two column joint analysis
    twoscalar_analyzer = enb.aanalysis.TwoNumericAnalyzer(
        csv_support_path=os.path.join(
            options.analysis_dir, "lossless_compression_analysis_twocolumn.csv"))
    twoscalar_analyzer.get_df(
        full_df=df,
        target_columns=column_pairs,
        column_to_properties=exp.joined_column_to_properties,
        group_by="task_label",
        selected_render_modes={"scatter"},
    )
